@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2017 Jean-Pierre Charras, jp.charras at wanadoo.fr
  * Copyright (C) 2013 CERN (www.cern.ch)
- * Copyright (C) 2004-2022 KiCad Developers, see change_log.txt for contributors.
+ * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -30,9 +30,13 @@
 #include "widgets/bitmap_button.h"
 
 #include <advanced_config.h>
+#include <background_jobs_monitor.h>
 #include <bitmaps.h>
 #include <build_version.h>
+#include <confirm.h>
 #include <dialogs/panel_kicad_launcher.h>
+#include <dialogs/dialog_update_check_prompt.h>
+#include <dialogs/panel_jobset.h>
 #include <eda_base_frame.h>
 #include <executable_names.h>
 #include <file_history.h>
@@ -44,6 +48,7 @@
 #include <kiway.h>
 #include <kiway_express.h>
 #include <launch_ext.h>
+#include <notifications_manager.h>
 #include <reporter.h>
 #include <project/project_local_settings.h>
 #include <sch_file_versions.h>
@@ -57,15 +62,16 @@
 #include <tools/kicad_manager_control.h>
 #include <wildcards_and_files_ext.h>
 #include <widgets/app_progress_dialog.h>
+#include <widgets/kistatusbar.h>
 #include <wx/ffile.h>
 #include <wx/filedlg.h>
-#include <wx/dcclient.h>
 #include <wx/dnd.h>
 #include <wx/process.h>
 #include <atomic>
+#include <update_manager.h>
+#include <jobs/jobset.h>
 
-
-#include <../pcbnew/plugins/kicad/pcb_plugin.h>   // for SEXPR_BOARD_FILE_VERSION def
+#include <../pcbnew/pcb_io/kicad_sexpr/pcb_io_kicad_sexpr.h>   // for SEXPR_BOARD_FILE_VERSION def
 
 
 #ifdef __WXMAC__
@@ -76,6 +82,11 @@
 #include "kicad_manager_frame.h"
 #include "settings/kicad_settings.h"
 
+#include <project/project_file.h>
+
+
+#define EDITORS_CAPTION _( "Editors" )
+#define PROJECT_FILES_CAPTION _( "Project Files" )
 
 #define SEP()   wxFileName::GetPathSeparator()
 
@@ -94,6 +105,8 @@ BEGIN_EVENT_TABLE( KICAD_MANAGER_FRAME, EDA_BASE_FRAME )
     EVT_MENU( ID_READ_ZIP_ARCHIVE, KICAD_MANAGER_FRAME::OnUnarchiveFiles )
     EVT_MENU( ID_IMPORT_CADSTAR_ARCHIVE_PROJECT, KICAD_MANAGER_FRAME::OnImportCadstarArchiveFiles )
     EVT_MENU( ID_IMPORT_EAGLE_PROJECT, KICAD_MANAGER_FRAME::OnImportEagleFiles )
+    EVT_MENU( ID_IMPORT_EASYEDA_PROJECT, KICAD_MANAGER_FRAME::OnImportEasyEdaFiles )
+    EVT_MENU( ID_IMPORT_EASYEDAPRO_PROJECT, KICAD_MANAGER_FRAME::OnImportEasyEdaProFiles )
 
     // Range menu events
     EVT_MENU_RANGE( ID_LANGUAGE_CHOICE, ID_LANGUAGE_CHOICE_END,
@@ -110,7 +123,6 @@ BEGIN_EVENT_TABLE( KICAD_MANAGER_FRAME, EDA_BASE_FRAME )
 
 END_EVENT_TABLE()
 
-
 // See below the purpose of this include
 #include <wx/xml/xml.h>
 
@@ -120,10 +132,13 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
                         KICAD_MANAGER_FRAME_NAME, &::Kiway, unityScale ),
         m_leftWin( nullptr ),
         m_launcher( nullptr ),
-        m_mainToolBar( nullptr )
+        m_mainToolBar( nullptr ),
+        m_lastToolbarIconSize( 0 )
 {
+    const int defaultLeftWinWidth = FromDIP( 250 );
+
     m_active_project = false;
-    m_leftWinWidth = 250;       // Default value
+    m_leftWinWidth = defaultLeftWinWidth; // Default value
     m_aboutTitle = "KiCad";
 
     // JPC: A very ugly hack to fix an issue on Linux: if the wxbase315u_xml_gcc_custom.so is
@@ -134,7 +149,14 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
 
     // Create the status line (bottom of the frame).  Left half is for project name; right half
     // is for Reporter (currently used by archiver/unarchiver and PCM).
-    CreateStatusBar( 2 );
+    // Note: this is a KISTATUSBAR status bar. Therefore the specified number of fields
+    // is the extra number of fields, not the full field count.
+    // We need here 2 fields: the extra fiels to display the project name, and another field
+    // to display a info (specific to Windows) using the FIELD_OFFSET_BGJOB_TEXT id offset (=1)
+    // So the extra field count is 1
+    CreateStatusBar( 1 );
+    Pgm().GetBackgroundJobMonitor().RegisterStatusBar( (KISTATUSBAR*) GetStatusBar() );
+    Pgm().GetNotificationsManager().RegisterStatusBar( (KISTATUSBAR*) GetStatusBar() );
     GetStatusBar()->SetFont( KIUI::GetStatusFont( this ) );
 
     // Give an icon
@@ -143,7 +165,11 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
 
     if( IsNightlyVersion())
     {
-        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad_nightly ) );
+        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad_nightly, 48 ) );
+        icon_bundle.AddIcon( icon );
+        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad_nightly, 128 ) );
+        icon_bundle.AddIcon( icon );
+        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad_nightly, 256 ) );
         icon_bundle.AddIcon( icon );
         icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad_nightly_32 ) );
         icon_bundle.AddIcon( icon );
@@ -152,7 +178,11 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
     }
     else
     {
-        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad ) );
+        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad, 48 ) );
+        icon_bundle.AddIcon( icon );
+        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad, 128 ) );
+        icon_bundle.AddIcon( icon );
+        icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad, 256 ) );
         icon_bundle.AddIcon( icon );
         icon.CopyFromBitmap( KiBitmap( BITMAPS::icon_kicad_32 ) );
         icon_bundle.AddIcon( icon );
@@ -167,25 +197,6 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
 
     m_pcmButton = nullptr;
     m_pcmUpdateCount = 0;
-    m_pcm = std::make_shared<PLUGIN_CONTENT_MANAGER>(
-            [this]( int aUpdateCount )
-            {
-                m_pcmUpdateCount = aUpdateCount;
-                CallAfter(
-                        [this]()
-                        {
-                            updatePcmButtonBadge();
-                        } );
-            },
-            [this]( const wxString aText )
-            {
-                CallAfter(
-                        [aText, this]()
-                        {
-                            SetStatusText( aText, 1 );
-                        } );
-            } );
-    m_pcm->SetRepositoryList( kicadSettings()->m_PcmRepositories );
 
     // Left window: is the box which display tree project
     m_leftWin = new PROJECT_TREE_PANE( this );
@@ -193,16 +204,14 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
     setupTools();
     setupUIConditions();
 
-    m_launcher = new PANEL_KICAD_LAUNCHER( this );
-
-    RecreateBaseHToolbar();
+    RecreateBaseLeftToolbar();
     ReCreateMenuBar();
 
     m_auimgr.SetManagedWindow( this );
     m_auimgr.SetFlags( wxAUI_MGR_LIVE_RESIZE );
 
-    m_auimgr.AddPane( m_mainToolBar, EDA_PANE().HToolbar().Name( "MainToolbar" ).Left()
-                      .Layer( 2 ) );
+    m_auimgr.AddPane( m_mainToolBar,
+                      EDA_PANE().VToolbar().Name( "MainToolbar" ).Left().Layer( 2 ) );
 
     // BestSize() does not always set the actual pane size of m_leftWin to the required value.
     // It happens when m_leftWin is too large (roughly > 1/3 of the kicad manager frame width.
@@ -210,23 +219,46 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
     // A trick is to use MinSize() to set the required pane width,
     // and after give a reasonable MinSize value
     m_auimgr.AddPane( m_leftWin, EDA_PANE().Palette().Name( "ProjectTree" ).Left().Layer( 1 )
-                      .Caption( _( "Project Files" ) ).PaneBorder( false )
+                      .Caption( PROJECT_FILES_CAPTION ).PaneBorder( false )
                       .MinSize( m_leftWinWidth, -1 ).BestSize( m_leftWinWidth, -1 ) );
 
-    m_auimgr.AddPane( m_launcher, EDA_PANE().Canvas().Name( "Launcher" ).Center()
-                      .Caption( _( "Editors" ) ).PaneBorder( false )
-                      .MinSize( m_launcher->GetBestSize() ) );
+    wxSize client_size = GetClientSize();
+    m_notebook = new wxAuiNotebook( this, wxID_ANY, wxPoint( client_size.x, client_size.y ),
+                                    FromDIP( wxSize( 700, 590 ) ),
+                                    wxAUI_NB_TOP | wxAUI_NB_CLOSE_ON_ALL_TABS | wxAUI_NB_TAB_MOVE
+                                            | wxAUI_NB_SCROLL_BUTTONS | wxNO_BORDER );
+
+    m_notebook->Bind( wxEVT_AUINOTEBOOK_PAGE_CLOSE,
+                      &KICAD_MANAGER_FRAME::onNotebookPageCloseRequest, this );
+    m_notebook->Bind( wxEVT_AUINOTEBOOK_PAGE_CLOSED,
+                      &KICAD_MANAGER_FRAME::onNotebookPageCountChanged, this );
+    m_launcher = new PANEL_KICAD_LAUNCHER( m_notebook );
+
+    m_notebook->Freeze();
+    m_launcher->SetClosable( false );
+    m_notebook->AddPage( m_launcher, EDITORS_CAPTION, false );
+    m_notebook->SetTabCtrlHeight( 0 );
+    m_notebook->Thaw();
+
+    m_auimgr.AddPane( m_notebook,
+                      EDA_PANE().Canvas().Name( "Editors" ).Center().Caption( EDITORS_CAPTION )
+                                .PaneBorder( false ).MinSize( m_notebook->GetBestSize() ) );
 
     m_auimgr.Update();
 
     // Now the actual m_leftWin size is set, give it a reasonable min width
-    m_auimgr.GetPane( m_leftWin ).MinSize( 250, -1 );
+    m_auimgr.GetPane( m_leftWin ).MinSize( defaultLeftWinWidth, -1 );
+
 
     wxSizer* mainSizer = GetSizer();
 
     // Only fit the initial window size the first time KiCad is run.
     if( mainSizer && config()->m_Window.state.size_x == 0 && config()->m_Window.state.size_y == 0 )
+    {
+        Layout();
         mainSizer->Fit( this );
+        Center();
+    }
 
     if( ADVANCED_CFG::GetCfg().m_HideVersionFromTitle )
         SetTitle( wxT( "KiCad" ) );
@@ -237,24 +269,14 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
     m_leftWin->SetFocus();
 
     // Init for dropping files
-    m_acceptedExts.emplace( ProjectFileExtension, &KICAD_MANAGER_ACTIONS::loadProject );
-    m_acceptedExts.emplace( LegacyProjectFileExtension, &KICAD_MANAGER_ACTIONS::loadProject );
+    m_acceptedExts.emplace( FILEEXT::ProjectFileExtension, &KICAD_MANAGER_ACTIONS::loadProject );
+    m_acceptedExts.emplace( FILEEXT::LegacyProjectFileExtension, &KICAD_MANAGER_ACTIONS::loadProject );
 
     // Gerber files
     // Note that all gerber files are aliased as GerberFileExtension
-    m_acceptedExts.emplace( GerberFileExtension, &KICAD_MANAGER_ACTIONS::viewDroppedGerbers );
-    m_acceptedExts.emplace( GerberJobFileExtension, &KICAD_MANAGER_ACTIONS::viewDroppedGerbers );
-    m_acceptedExts.emplace( DrillFileExtension, &KICAD_MANAGER_ACTIONS::viewDroppedGerbers );
-
-    // Eagle files import
-    m_acceptedExts.emplace( EagleSchematicFileExtension,
-                            &KICAD_MANAGER_ACTIONS::importNonKicadProj );
-    m_acceptedExts.emplace( EaglePcbFileExtension, &KICAD_MANAGER_ACTIONS::importNonKicadProj );
-
-    // Cadstar files import
-    m_acceptedExts.emplace( CadstarSchematicFileExtension,
-                            &KICAD_MANAGER_ACTIONS::importNonKicadProj );
-    m_acceptedExts.emplace( CadstarPcbFileExtension, &KICAD_MANAGER_ACTIONS::importNonKicadProj );
+    m_acceptedExts.emplace( FILEEXT::GerberFileExtension, &KICAD_MANAGER_ACTIONS::viewDroppedGerbers );
+    m_acceptedExts.emplace( FILEEXT::GerberJobFileExtension, &KICAD_MANAGER_ACTIONS::viewDroppedGerbers );
+    m_acceptedExts.emplace( FILEEXT::DrillFileExtension, &KICAD_MANAGER_ACTIONS::viewDroppedGerbers );
 
     DragAcceptFiles( true );
 
@@ -265,11 +287,23 @@ KICAD_MANAGER_FRAME::KICAD_MANAGER_FRAME( wxWindow* parent, const wxString& titl
 
 KICAD_MANAGER_FRAME::~KICAD_MANAGER_FRAME()
 {
+    Unbind( wxEVT_CHAR, &TOOL_DISPATCHER::DispatchWxEvent, m_toolDispatcher );
+    Unbind( wxEVT_CHAR_HOOK, &TOOL_DISPATCHER::DispatchWxEvent, m_toolDispatcher );
+
+    m_notebook->Unbind( wxEVT_AUINOTEBOOK_PAGE_CLOSE,
+                        &KICAD_MANAGER_FRAME::onNotebookPageCloseRequest, this );
+    m_notebook->Unbind( wxEVT_AUINOTEBOOK_PAGE_CLOSED,
+                        &KICAD_MANAGER_FRAME::onNotebookPageCountChanged, this );
+
+    Pgm().GetBackgroundJobMonitor().UnregisterStatusBar( (KISTATUSBAR*) GetStatusBar() );
+    Pgm().GetNotificationsManager().UnregisterStatusBar( (KISTATUSBAR*) GetStatusBar() );
+
     // Shutdown all running tools
     if( m_toolManager )
         m_toolManager->ShutdownAllTools();
 
-    m_pcm->StopBackgroundUpdate();
+    if( m_pcm )
+        m_pcm->StopBackgroundUpdate();
 
     delete m_actions;
     delete m_toolManager;
@@ -278,6 +312,88 @@ KICAD_MANAGER_FRAME::~KICAD_MANAGER_FRAME()
     m_auimgr.UnInit();
 }
 
+void KICAD_MANAGER_FRAME::HideTabsIfNeeded()
+{
+    if( m_notebook->GetPageCount() == 1 )
+        m_notebook->SetTabCtrlHeight( 0 );
+    else
+        m_notebook->SetTabCtrlHeight( -1 );
+}
+
+
+void KICAD_MANAGER_FRAME::onNotebookPageCountChanged( wxAuiNotebookEvent& evt )
+{
+    HideTabsIfNeeded();
+}
+
+
+void KICAD_MANAGER_FRAME::onNotebookPageCloseRequest( wxAuiNotebookEvent& evt )
+{
+    wxAuiNotebook* notebook = (wxAuiNotebook*) evt.GetEventObject();
+    wxWindow*      page = notebook->GetPage( evt.GetSelection() );
+
+    if( PANEL_NOTEBOOK_BASE* panel = dynamic_cast<PANEL_NOTEBOOK_BASE*>( page ) )
+    {
+        if( panel->GetClosable() )
+        {
+            if( !panel->GetCanClose() )
+                evt.Veto();
+
+            CallAfter(
+                    [this]()
+                    {
+                        SaveOpenJobSetsToLocalSettings();
+                    } );
+        }
+        else
+        {
+            evt.Veto();
+        }
+    }
+}
+
+
+wxStatusBar* KICAD_MANAGER_FRAME::OnCreateStatusBar( int number, long style, wxWindowID id,
+                                                     const wxString& name )
+{
+    return new KISTATUSBAR( number, this, id );
+}
+
+
+void KICAD_MANAGER_FRAME::CreatePCM()
+{
+    // creates the PLUGIN_CONTENT_MANAGER, if not exists
+    if( m_pcm )
+        return;
+
+    m_pcm = std::make_shared<PLUGIN_CONTENT_MANAGER>(
+            [this]( int aUpdateCount )
+            {
+                m_pcmUpdateCount = aUpdateCount;
+
+                if( aUpdateCount > 0 )
+                {
+                    Pgm().GetNotificationsManager().CreateOrUpdate(
+                            wxS( "pcm" ),
+                            _( "PCM Updates Available" ),
+                            wxString::Format( _( "%d package update(s) avaliable" ), aUpdateCount ),
+                            wxT( "" ) );
+                }
+                else
+                {
+                    Pgm().GetNotificationsManager().Remove( wxS( "pcm" ) );
+                }
+
+                CallAfter(
+                        [this]()
+                        {
+                            updatePcmButtonBadge();
+                        } );
+            });
+
+    m_pcm->SetRepositoryList( kicadSettings()->m_PcmRepositories );
+}
+
 
 void KICAD_MANAGER_FRAME::setupTools()
 {
@@ -308,10 +424,10 @@ void KICAD_MANAGER_FRAME::setupUIConditions()
     wxASSERT( manager );
 
     auto activeProject =
-        [this] ( const SELECTION& )
-        {
-            return m_active_project;
-        };
+            [this] ( const SELECTION& )
+            {
+                return m_active_project;
+            };
 
 #define ENABLE( x ) ACTION_CONDITIONS().Enable( x )
 
@@ -320,6 +436,8 @@ void KICAD_MANAGER_FRAME::setupUIConditions()
 
     manager->SetConditions( ACTIONS::saveAs,                       activeProjectCond );
     manager->SetConditions( KICAD_MANAGER_ACTIONS::closeProject,   activeProjectCond );
+    manager->SetConditions( KICAD_MANAGER_ACTIONS::newJobsetFile,  activeProjectCond );
+    manager->SetConditions( KICAD_MANAGER_ACTIONS::openJobsetFile, activeProjectCond );
 
     // These are just here for text boxes, search boxes, etc. in places such as the standard
     // file dialogs.
@@ -367,7 +485,7 @@ const wxString KICAD_MANAGER_FRAME::SchFileName()
 {
    wxFileName   fn( GetProjectFileName() );
 
-   fn.SetExt( KiCadSchematicFileExtension );
+   fn.SetExt( FILEEXT::KiCadSchematicFileExtension );
    return fn.GetFullPath();
 }
 
@@ -376,7 +494,7 @@ const wxString KICAD_MANAGER_FRAME::SchLegacyFileName()
 {
    wxFileName   fn( GetProjectFileName() );
 
-   fn.SetExt( LegacySchematicFileExtension );
+   fn.SetExt( FILEEXT::LegacySchematicFileExtension );
    return fn.GetFullPath();
 }
 
@@ -385,7 +503,7 @@ const wxString KICAD_MANAGER_FRAME::PcbFileName()
 {
    wxFileName   fn( GetProjectFileName() );
 
-   fn.SetExt( PcbFileExtension );
+   fn.SetExt( FILEEXT::PcbFileExtension );
    return fn.GetFullPath();
 }
 
@@ -394,7 +512,7 @@ const wxString KICAD_MANAGER_FRAME::PcbLegacyFileName()
 {
    wxFileName   fn( GetProjectFileName() );
 
-   fn.SetExt( LegacyPcbFileExtension );
+   fn.SetExt( FILEEXT::LegacyPcbFileExtension );
    return fn.GetFullPath();
 }
 
@@ -424,6 +542,11 @@ void KICAD_MANAGER_FRAME::OnSize( wxSizeEvent& event )
 
     PrintPrjInfo();
 
+#if defined( _WIN32 )
+    KISTATUSBAR* statusBar = static_cast<KISTATUSBAR*>( GetStatusBar() );
+    statusBar->SetEllipsedTextField( m_FileWatcherInfo, 1 );
+#endif
+
     event.Skip();
 }
 
@@ -437,10 +560,10 @@ void KICAD_MANAGER_FRAME::DoWithAcceptedFiles()
     {
         wxString ext = fileName.GetExt();
 
-        if( ext == ProjectFileExtension || ext == LegacyProjectFileExtension )
+        if( ext == FILEEXT::ProjectFileExtension || ext == FILEEXT::LegacyProjectFileExtension )
         {
             wxString fn = fileName.GetFullPath();
-            m_toolManager->RunAction( *m_acceptedExts.at( fileName.GetExt() ), true, &fn );
+            m_toolManager->RunAction<wxString*>( *m_acceptedExts.at( fileName.GetExt() ), &fn );
 
             return;
         }
@@ -454,8 +577,8 @@ void KICAD_MANAGER_FRAME::DoWithAcceptedFiles()
     {
         wxString ext = fileName.GetExt();
 
-        if( ext == GerberJobFileExtension || ext == DrillFileExtension
-            || IsGerberFileExtension( ext ) )
+        if( ext == FILEEXT::GerberJobFileExtension || ext == FILEEXT::DrillFileExtension
+            || FILEEXT::IsGerberFileExtension( ext ) )
         {
             gerberFiles += wxT( '\"' );
             gerberFiles += fileName.GetFullPath() + wxT( '\"' );
@@ -464,7 +587,7 @@ void KICAD_MANAGER_FRAME::DoWithAcceptedFiles()
         else
         {
             wxString fn = fileName.GetFullPath();
-            m_toolManager->RunAction( *m_acceptedExts.at( fileName.GetExt() ), true, &fn );
+            m_toolManager->RunAction<wxString*>( *m_acceptedExts.at( fileName.GetExt() ), &fn );
         }
     }
 
@@ -476,8 +599,8 @@ void KICAD_MANAGER_FRAME::DoWithAcceptedFiles()
         if( wxFileExists( fullEditorName ) )
         {
             wxString command = fullEditorName + " " + gerberFiles;
-            m_toolManager->RunAction( *m_acceptedExts.at( GerberFileExtension ),
-                                      true, &command );
+            m_toolManager->RunAction<wxString*>( *m_acceptedExts.at( FILEEXT::GerberFileExtension ),
+                                                 &command );
         }
     }
 }
@@ -488,6 +611,17 @@ bool KICAD_MANAGER_FRAME::canCloseWindow( wxCloseEvent& aEvent )
     KICAD_SETTINGS* settings = kicadSettings();
     settings->m_OpenProjects = GetSettingsManager()->GetOpenProjects();
 
+    for( size_t i = 0; i < m_notebook->GetPageCount(); i++ )
+    {
+        wxWindow* page = m_notebook->GetPage( i );
+
+        if( PANEL_NOTEBOOK_BASE* panel = dynamic_cast<PANEL_NOTEBOOK_BASE*>( page ) )
+        {
+            if( !panel->GetCanClose() )
+                return false;
+        }
+    }
+
     // CloseProject will recursively ask all the open editors if they need to save changes.
     // If any of them cancel then we need to cancel closing the KICAD_MANAGER_FRAME.
     if( CloseProject( true ) )
@@ -525,6 +659,14 @@ void KICAD_MANAGER_FRAME::doCloseWindow()
     }
 #endif
 
+    for( size_t i = 0; i < m_notebook->GetPageCount(); i++ )
+    {
+        wxWindow* page = m_notebook->GetPage( i );
+
+        if( dynamic_cast<PANEL_NOTEBOOK_BASE*>( page ) )
+            m_notebook->DeletePage( i );
+    }
+
     m_leftWin->Show( false );
     Pgm().m_Quitting = true;
 
@@ -536,6 +678,29 @@ void KICAD_MANAGER_FRAME::doCloseWindow()
 }
 
 
+void KICAD_MANAGER_FRAME::SaveOpenJobSetsToLocalSettings( bool aIsExplicitUserSave )
+{
+    PROJECT_LOCAL_SETTINGS& cfg = Prj().GetLocalSettings();
+
+    if( !aIsExplicitUserSave && !cfg.ShouldAutoSave() )
+        return;
+
+    cfg.m_OpenJobSets.clear();
+
+    for( size_t i = 0; i < m_notebook->GetPageCount(); i++ )
+    {
+        if( PANEL_JOBSET* jobset = dynamic_cast<PANEL_JOBSET*>( m_notebook->GetPage( i ) ) )
+        {
+            wxFileName jobsetFn( jobset->GetFilePath() );
+            jobsetFn.MakeRelativeTo( Prj().GetProjectPath() );
+            cfg.m_OpenJobSets.emplace_back( jobsetFn.GetFullPath() );
+        }
+    }
+
+    cfg.SaveToFile( Prj().GetProjectPath() );
+}
+
+
 void KICAD_MANAGER_FRAME::OnExit( wxCommandEvent& event )
 {
     Close( true );
@@ -547,15 +712,21 @@ bool KICAD_MANAGER_FRAME::CloseProject( bool aSave )
     if( !Kiway().PlayersClose( false ) )
         return false;
 
+    bool shouldSaveProject = Prj().GetLocalSettings().ShouldAutoSave()
+                             && Prj().GetProjectFile().ShouldAutoSave();
+
     // Save the project file for the currently loaded project.
     if( m_active_project )
     {
         SETTINGS_MANAGER& mgr = Pgm().GetSettingsManager();
 
-        mgr.TriggerBackupIfNeeded( NULL_REPORTER::GetInstance() );
+        if( shouldSaveProject )
+        {
+            mgr.TriggerBackupIfNeeded( NULL_REPORTER::GetInstance() );
 
-        if( aSave )
-            mgr.SaveProject();
+            if( aSave )
+                mgr.SaveProject();
+        }
 
         m_active_project = false;
         mgr.UnloadProject( &Prj() );
@@ -563,12 +734,63 @@ bool KICAD_MANAGER_FRAME::CloseProject( bool aSave )
 
     SetStatusText( "" );
 
+    // Traverse pages in reverse order so deleting them doesn't mess up our iterator.
+    for( int i = (int) m_notebook->GetPageCount() - 1; i >= 0; i-- )
+    {
+        wxWindow* page = m_notebook->GetPage( i );
+
+        if( PANEL_NOTEBOOK_BASE* panel = dynamic_cast<PANEL_NOTEBOOK_BASE*>( page ) )
+        {
+            if( panel->GetProjectTied() )
+                m_notebook->DeletePage( i );
+        }
+    }
+
     m_leftWin->EmptyTreePrj();
+    HideTabsIfNeeded();
 
     return true;
 }
 
 
+void KICAD_MANAGER_FRAME::OpenJobsFile( const wxFileName& aFileName, bool aCreate,
+                                        bool aResaveProjectPreferences )
+{
+    for( size_t i = 0; i < m_notebook->GetPageCount(); i++ )
+    {
+        if( PANEL_JOBSET* panel = dynamic_cast<PANEL_JOBSET*>( m_notebook->GetPage( i ) ) )
+        {
+            if( aFileName.GetFullPath() == panel->GetFilePath() )
+            {
+                m_notebook->SetSelection( i );
+                return;
+            }
+        }
+    }
+
+    try
+    {
+        std::unique_ptr<JOBSET> jobsFile =
+                std::make_unique<JOBSET>( aFileName.GetFullPath().ToStdString() );
+
+        jobsFile->LoadFromFile();
+
+        PANEL_JOBSET* jobPanel = new PANEL_JOBSET( m_notebook, this, std::move( jobsFile ) );
+        jobPanel->SetProjectTied( true );
+        jobPanel->SetClosable( true );
+        m_notebook->AddPage( jobPanel, aFileName.GetFullName(), true );
+        HideTabsIfNeeded();
+
+        if( aResaveProjectPreferences )
+            SaveOpenJobSetsToLocalSettings();
+    }
+    catch( ... )
+    {
+        DisplayErrorMessage( this, _( "Error opening jobs file" ) );
+    }
+}
+
+
 void KICAD_MANAGER_FRAME::LoadProject( const wxFileName& aProjectFileName )
 {
     // The project file should be valid by the time we get here or something has gone wrong.
@@ -590,10 +812,25 @@ void KICAD_MANAGER_FRAME::LoadProject( const wxFileName& aProjectFileName )
     if( aProjectFileName.IsDirWritable() )
         SetMruPath( Prj().GetProjectPath() ); // Only set MRU path if we have write access. Why?
 
-    UpdateFileHistory( Prj().GetProjectFullName() );
+    // Save history & window state to disk now.  Don't wait around for a crash.
+    KICAD_SETTINGS* settings = kicadSettings();
+    SaveSettings( settings );
+    settings->SaveToFile( Pgm().GetSettingsManager().GetPathForSettingsFile( settings ) );
 
     m_leftWin->ReCreateTreePrj();
 
+    for( const wxString& jobset : Prj().GetLocalSettings().m_OpenJobSets )
+    {
+        wxFileName jobsetFn( jobset );
+        jobsetFn.MakeAbsolute( Prj().GetProjectPath() );
+
+        if( jobsetFn.Exists() )
+            OpenJobsFile( jobsetFn.GetFullPath(), false, false );
+    }
+
+    // Always start with the apps page
+    m_notebook->SetSelection( 0 );
+
     // Rebuild the list of watched paths.
     // however this is possible only when the main loop event handler is running,
     // so we use it to run the rebuild function.
@@ -618,7 +855,7 @@ void KICAD_MANAGER_FRAME::CreateNewProject( const wxFileName& aProjectFileName,
     if( !aProjectFileName.FileExists() )
     {
         wxFileName legacyPro( aProjectFileName );
-        legacyPro.SetExt( LegacyProjectFileExtension );
+        legacyPro.SetExt( FILEEXT::LegacyProjectFileExtension );
 
         if( legacyPro.FileExists() )
         {
@@ -633,7 +870,7 @@ void KICAD_MANAGER_FRAME::CreateNewProject( const wxFileName& aProjectFileName,
             wxString srcFileName = sys_search().FindValidPath( "kicad.kicad_pro" );
 
             wxFileName destFileName( aProjectFileName );
-            destFileName.SetExt( ProjectFileExtension );
+            destFileName.SetExt( FILEEXT::ProjectFileExtension );
 
             // Create a minimal project file if the template project file could not be copied
             if( !wxFileName::FileExists( srcFileName )
@@ -655,7 +892,7 @@ void KICAD_MANAGER_FRAME::CreateNewProject( const wxFileName& aProjectFileName,
     if( aCreateStubFiles )
     {
         wxFileName fn( aProjectFileName.GetFullPath() );
-        fn.SetExt( KiCadSchematicFileExtension );
+        fn.SetExt( FILEEXT::KiCadSchematicFileExtension );
 
         // If a <project>.kicad_sch file does not exist, create a "stub" file ( minimal schematic
         // file ).
@@ -664,19 +901,19 @@ void KICAD_MANAGER_FRAME::CreateNewProject( const wxFileName& aProjectFileName,
             wxFFile file( fn.GetFullPath(), "wb" );
 
             if( file.IsOpened() )
-                file.Write( wxString::Format( "(kicad_sch (version %d) (generator eeschema)\n"
+                file.Write( wxString::Format( "(kicad_sch (version %d) (generator \"eeschema\") (generator_version \"%s\")\n"
                                               "  (paper \"A4\")\n  (lib_symbols)\n"
                                               "  (symbol_instances)\n)\n",
-                                              SEXPR_SCHEMATIC_FILE_VERSION ) );
+                                              SEXPR_SCHEMATIC_FILE_VERSION, GetMajorMinorVersion() ) );
 
             // wxFFile dtor will close the file
         }
 
         // If a <project>.kicad_pcb or <project>.brd file does not exist,
         // create a .kicad_pcb "stub" file
-        fn.SetExt( KiCadPcbFileExtension );
+        fn.SetExt( FILEEXT::KiCadPcbFileExtension );
         wxFileName leg_fn( fn );
-        leg_fn.SetExt( LegacyPcbFileExtension );
+        leg_fn.SetExt( FILEEXT::LegacyPcbFileExtension );
 
         if( !fn.FileExists() && !leg_fn.FileExists() )
         {
@@ -684,14 +921,17 @@ void KICAD_MANAGER_FRAME::CreateNewProject( const wxFileName& aProjectFileName,
 
             if( file.IsOpened() )
                 // Create a small dummy file as a stub for pcbnew:
-                file.Write( wxString::Format( "(kicad_pcb (version %d) (generator pcbnew)\n)",
-                                              SEXPR_BOARD_FILE_VERSION ) );
+                file.Write( wxString::Format( "(kicad_pcb (version %d) (generator \"pcbnew\") (generator_version \"%s\")\n)",
+                                              SEXPR_BOARD_FILE_VERSION, GetMajorMinorVersion() ) );
 
             // wxFFile dtor will close the file
         }
     }
 
-    UpdateFileHistory( aProjectFileName.GetFullPath() );
+    // Save history & window state to disk now.  Don't wait around for a crash.
+    KICAD_SETTINGS* settings = kicadSettings();
+    SaveSettings( settings );
+    settings->SaveToFile( Pgm().GetSettingsManager().GetPathForSettingsFile( settings ) );
 
     m_openSavedWindows = true;
 }
@@ -700,12 +940,12 @@ void KICAD_MANAGER_FRAME::CreateNewProject( const wxFileName& aProjectFileName,
 void KICAD_MANAGER_FRAME::OnOpenFileInTextEditor( wxCommandEvent& event )
 {
     // show all files in file dialog (in Kicad all files are editable texts):
-    wxString wildcard = AllFilesWildcard();
+    wxString wildcard = FILEEXT::AllFilesWildcard();
 
     wxString default_dir = Prj().GetProjectPath();
 
-    wxFileDialog dlg( this, _( "Load File to Edit" ), default_dir,
-                      wxEmptyString, wildcard, wxFD_OPEN );
+    wxFileDialog dlg( this, _( "Edit File in Text Editor" ), default_dir,  wxEmptyString, wildcard,
+                      wxFD_OPEN );
 
     if( dlg.ShowModal() == wxID_CANCEL )
         return;
@@ -713,7 +953,7 @@ void KICAD_MANAGER_FRAME::OnOpenFileInTextEditor( wxCommandEvent& event )
     wxString filename = dlg.GetPath();
 
     if( !dlg.GetPath().IsEmpty() && !Pgm().GetTextEditor().IsEmpty() )
-        m_toolManager->RunAction( KICAD_MANAGER_ACTIONS::openTextEditor, true, &filename );
+        m_toolManager->RunAction<wxString*>( KICAD_MANAGER_ACTIONS::openTextEditor, &filename );
 }
 
 
@@ -743,20 +983,38 @@ void KICAD_MANAGER_FRAME::ShowChangedLanguage()
     EDA_BASE_FRAME::ShowChangedLanguage();
 
     // tooltips in toolbars
-    RecreateBaseHToolbar();
+    RecreateBaseLeftToolbar();
     m_launcher->CreateLaunchers();
 
+    // update captions
+    int pageId = m_notebook->FindPage( m_launcher );
+
+    if( pageId != wxNOT_FOUND )
+        m_notebook->SetPageText( pageId, EDITORS_CAPTION );
+
+    m_auimgr.GetPane( m_leftWin ).Caption( PROJECT_FILES_CAPTION );
+    m_auimgr.Update();
+
+    m_leftWin->FileWatcherReset();
+
     PrintPrjInfo();
 }
 
 
-void KICAD_MANAGER_FRAME::CommonSettingsChanged( bool aEnvVarsChanged, bool aTextVarsChanged )
+void KICAD_MANAGER_FRAME::CommonSettingsChanged( int aFlags )
 {
-    EDA_BASE_FRAME::CommonSettingsChanged( aEnvVarsChanged, aTextVarsChanged );
+    EDA_BASE_FRAME::CommonSettingsChanged( aFlags );
 
-    if( aEnvVarsChanged )
-    {
+    if( m_pcm && ( aFlags & ENVVARS_CHANGED ) )
         m_pcm->ReadEnvVar();
+
+    COMMON_SETTINGS* settings = Pgm().GetCommonSettings();
+
+    if( m_lastToolbarIconSize == 0
+        || m_lastToolbarIconSize != settings->m_Appearance.toolbar_icon_size )
+    {
+        onToolbarSizeChanged();
+        m_lastToolbarIconSize = settings->m_Appearance.toolbar_icon_size;
     }
 }
 
@@ -810,6 +1068,9 @@ void KICAD_MANAGER_FRAME::SaveSettings( APP_SETTINGS_BASE* aCfg )
     wxCHECK( settings, /*void*/);
 
     settings->m_LeftWinWidth = m_leftWin->GetSize().x;
+
+    if( !m_isClosing )
+        settings->m_OpenProjects = GetSettingsManager()->GetOpenProjects();
 }
 
 
@@ -818,16 +1079,8 @@ void KICAD_MANAGER_FRAME::PrintPrjInfo()
     // wxStatusBar's wxELLIPSIZE_MIDDLE flag doesn't work (at least on Mac).
 
     wxString     status = wxString::Format( _( "Project: %s" ), Prj().GetProjectFullName() );
-    wxStatusBar* statusBar = GetStatusBar();
-    int          width = statusBar->GetSize().GetWidth() / 2;
-
-    if( width > 20 )
-    {
-        wxClientDC dc( this );
-        status = wxControl::Ellipsize( status, dc, wxELLIPSIZE_MIDDLE, width );
-    }
-
-    SetStatusText( status );
+    KISTATUSBAR* statusBar = static_cast<KISTATUSBAR*>( GetStatusBar() );
+    statusBar->SetEllipsedTextField( status, 0 );
 }
 
 
@@ -856,7 +1109,7 @@ void KICAD_MANAGER_FRAME::OnIdle( wxIdleEvent& aEvent )
                                Prj().GetLocalSettings().m_files.end(),
                                [&]( const PROJECT_FILE_STATE& f )
                                {
-                                   return !f.fileName.EndsWith( ProjectFileExtension ) && f.open;
+                                   return !f.fileName.EndsWith( FILEEXT::ProjectFileExtension ) && f.open;
                                } );
 
         if( previousOpenCount > 0 )
@@ -879,15 +1132,15 @@ void KICAD_MANAGER_FRAME::OnIdle( wxIdleEvent& aEvent )
                     openedFiles.insert( file.fileName );
                     wxFileName fn( file.fileName );
 
-                    if( fn.GetExt() == LegacySchematicFileExtension
-                            || fn.GetExt() == KiCadSchematicFileExtension )
+                    if( fn.GetExt() == FILEEXT::LegacySchematicFileExtension
+                        || fn.GetExt() == FILEEXT::KiCadSchematicFileExtension )
                     {
-                        GetToolManager()->RunAction( KICAD_MANAGER_ACTIONS::editSchematic, true );
+                        GetToolManager()->RunAction( KICAD_MANAGER_ACTIONS::editSchematic );
                     }
-                    else if( fn.GetExt() == LegacyPcbFileExtension
-                             || fn.GetExt() == KiCadPcbFileExtension )
+                    else if( fn.GetExt() == FILEEXT::LegacyPcbFileExtension
+                             || fn.GetExt() == FILEEXT::KiCadPcbFileExtension )
                     {
-                        GetToolManager()->RunAction( KICAD_MANAGER_ACTIONS::editPCB, true );
+                        GetToolManager()->RunAction( KICAD_MANAGER_ACTIONS::editPCB );
                     }
                 }
 
@@ -901,27 +1154,30 @@ void KICAD_MANAGER_FRAME::OnIdle( wxIdleEvent& aEvent )
 
     KICAD_SETTINGS* settings = kicadSettings();
 
-    if( settings->m_updateCheck == KICAD_SETTINGS::UPDATE_CHECK::UNINITIALIZED )
+    if( !Pgm().GetCommonSettings()->m_DoNotShowAgain.update_check_prompt )
     {
-        if( wxMessageBox( _( "Would you like to automatically check for plugin updates on startup?" ),
-                          _( "Check for updates" ), wxICON_QUESTION | wxYES_NO, this )
-            == wxYES )
-        {
-            settings->m_updateCheck = KICAD_SETTINGS::UPDATE_CHECK::ALLOWED;
-            settings->m_PcmUpdateCheck = true;
-        }
-        else
-        {
-            settings->m_updateCheck = KICAD_SETTINGS::UPDATE_CHECK::NOT_ALLOWED;
-            settings->m_PcmUpdateCheck = false;
-        }
+        auto prompt = new DIALOG_UPDATE_CHECK_PROMPT( this );
+        prompt->ShowModal();
+
+        Pgm().GetCommonSettings()->m_DoNotShowAgain.update_check_prompt = true;
     }
 
     if( KIPLATFORM::POLICY::GetPolicyBool( POLICY_KEY_PCM ) != KIPLATFORM::POLICY::PBOOL::DISABLED
         && settings->m_PcmUpdateCheck )
     {
+        if( !m_pcm )
+            CreatePCM();
+
         m_pcm->RunBackgroundUpdate();
     }
+
+#ifdef KICAD_UPDATE_CHECK
+    if( !m_updateManager && settings->m_KiCadUpdateCheck )
+    {
+        m_updateManager = std::make_unique<UPDATE_MANAGER>();
+        m_updateManager->CheckForUpdate( this );
+    }
+#endif
 }
 
 
@@ -950,3 +1206,18 @@ void KICAD_MANAGER_FRAME::updatePcmButtonBadge()
         m_pcmButton->Refresh();
     }
 }
+
+
+void KICAD_MANAGER_FRAME::onToolbarSizeChanged()
+{
+    // No idea why, but the same mechanism used in EDA_DRAW_FRAME doesn't work here
+    // the only thing that seems to work is to blow it all up and start from scratch.
+    m_auimgr.DetachPane( m_mainToolBar );
+    delete m_mainToolBar;
+    m_mainToolBar = nullptr;
+    RecreateBaseLeftToolbar();
+    m_auimgr.AddPane( m_mainToolBar, EDA_PANE().HToolbar().Name( "MainToolbar" ).Left()
+                      .Layer( 2 ) );
+
+    m_auimgr.Update();
+}
