@@ -2,7 +2,7 @@
  * This program source code file is part of KiCad, a free EDA CAD application.
  *
  * Copyright (C) 2016 CERN
- * Copyright (C) 2020-2023 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
  * @author Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
  *
  * This program is free software; you can redistribute it and/or
@@ -26,6 +26,7 @@
 #include <macros.h>
 #include <board.h>
 #include <footprint.h>
+#include <lset.h>
 #include <pcb_group.h>
 #include <pcb_track.h>
 #include <tool/tool_manager.h>
@@ -64,7 +65,7 @@ BOARD_COMMIT::BOARD_COMMIT( EDA_DRAW_FRAME* aFrame ) :
 
 
 BOARD_COMMIT::BOARD_COMMIT( TOOL_MANAGER* aMgr ) :
-        m_toolMgr( aMgr ),
+    m_toolMgr( aMgr ),
         m_isBoardEditor( false ),
         m_isFootprintEditor( false )
 {
@@ -76,6 +77,13 @@ BOARD_COMMIT::BOARD_COMMIT( TOOL_MANAGER* aMgr ) :
         m_isFootprintEditor = true;
 }
 
+BOARD_COMMIT::BOARD_COMMIT( TOOL_MANAGER* aMgr, bool aIsBoardEditor ) :
+    m_toolMgr( aMgr ),
+    m_isBoardEditor( aIsBoardEditor ),
+    m_isFootprintEditor( false )
+{
+}
+
 
 BOARD* BOARD_COMMIT::GetBoard() const
 {
@@ -85,8 +93,6 @@ BOARD* BOARD_COMMIT::GetBoard() const
 
 COMMIT& BOARD_COMMIT::Stage( EDA_ITEM* aItem, CHANGE_TYPE aChangeType, BASE_SCREEN* aScreen )
 {
-    wxCHECK( aItem, *this );
-
     // Many operations (move, rotate, etc.) are applied directly to a group's children, so they
     // must be staged as well.
     if( aChangeType == CHT_MODIFY )
@@ -201,9 +207,12 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
 
     for( COMMIT_LINE& ent : m_changes )
     {
-        BOARD_ITEM* boardItem = dynamic_cast<BOARD_ITEM*>( ent.m_item );
+        if( !ent.m_item || !ent.m_item->IsBOARD_ITEM() )
+            continue;
 
-        if( m_isBoardEditor && boardItem )
+        BOARD_ITEM* boardItem = static_cast<BOARD_ITEM*>( ent.m_item );
+
+        if( m_isBoardEditor )
         {
             if( boardItem->Type() == PCB_VIA_T || boardItem->Type() == PCB_FOOTPRINT_T
                     || boardItem->IsOnLayer( F_Mask ) || boardItem->IsOnLayer( B_Mask ) )
@@ -242,7 +251,7 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
             }
         }
 
-        if( boardItem && boardItem->IsSelected() )
+        if( boardItem->IsSelected() )
             selectedModified = true;
     }
 
@@ -252,12 +261,12 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
 
     for( COMMIT_LINE& ent : m_changes )
     {
-        int changeType = ent.m_type & CHT_TYPE;
-        int changeFlags = ent.m_type & CHT_FLAGS;
-        BOARD_ITEM* boardItem = dynamic_cast<BOARD_ITEM*>( ent.m_item );
+        if( !ent.m_item || !ent.m_item->IsBOARD_ITEM() )
+            continue;
 
-        wxASSERT( ent.m_item );
-        wxCHECK2( boardItem, continue );
+        BOARD_ITEM* boardItem = static_cast<BOARD_ITEM*>( ent.m_item );
+        int         changeType = ent.m_type & CHT_TYPE;
+        int         changeFlags = ent.m_type & CHT_FLAGS;
 
         switch( changeType )
         {
@@ -291,10 +300,10 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
                 }
             }
 
-            if( boardItem->Type() == PCB_GROUP_T )
+            if( boardItem->Type() == PCB_GROUP_T || boardItem->Type() == PCB_GENERATOR_T )
                 addedGroup = static_cast<PCB_GROUP*>( boardItem );
 
-            if( autofillZones && boardItem->Type() != PCB_MARKER_T )
+            if( m_isBoardEditor && autofillZones && boardItem->Type() != PCB_MARKER_T )
                 dirtyIntersectingZones( boardItem, changeType );
 
             if( view && boardItem->Type() != PCB_NETINFO_T )
@@ -324,7 +333,7 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
             if( parentFP && !( parentFP->GetFlags() & STRUCT_DELETED ) )
                 ent.m_parent = parentFP->m_Uuid;
 
-            if( autofillZones )
+            if( m_isBoardEditor && autofillZones && boardItem->Type() != PCB_MARKER_T )
                 dirtyIntersectingZones( boardItem, changeType );
 
             switch( boardItem->Type() )
@@ -338,7 +347,8 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
             case PCB_SHAPE_T:            // a shape (normally not on copper layers)
             case PCB_REFERENCE_IMAGE_T:  // a bitmap on an associated layer
             case PCB_GENERATOR_T:        // a generator on a layer
-            case PCB_TEXTBOX_T:          // a wrapped text on a layer
+            case PCB_TEXTBOX_T:          // a line-wrapped (and optionally bordered) text item
+            case PCB_TABLE_T:            // rows and columns of tablecells
             case PCB_TRACE_T:            // a track segment (segment on a copper layer)
             case PCB_ARC_T:              // an arced track segment (segment on a copper layer)
             case PCB_VIA_T:              // a via (like track segment on a copper layer)
@@ -351,24 +361,6 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
             case PCB_MARKER_T:           // a marker used to show something
             case PCB_ZONE_T:
             case PCB_FOOTPRINT_T:
-                if( view )
-                    view->Remove( boardItem );
-
-                if( !( changeFlags & CHT_DONE ) )
-                {
-                    if( parentFP )
-                    {
-                        parentFP->Remove( boardItem );
-                    }
-                    else
-                    {
-                        board->Remove( boardItem, REMOVE_MODE::BULK );
-                        bulkRemovedItems.push_back( boardItem );
-                    }
-                }
-
-                break;
-
             case PCB_GROUP_T:
                 if( view )
                     view->Remove( boardItem );
@@ -411,7 +403,7 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
                 if( !( aCommitFlags & SKIP_UNDO ) )
                 {
                     ITEM_PICKER itemWrapper( nullptr, boardItem, UNDO_REDO::UNGROUP );
-                    itemWrapper.SetLink( MakeImage( group ) );
+                    itemWrapper.SetGroupId( group->m_Uuid );
                     undoList.PushItem( itemWrapper );
                 }
 
@@ -433,7 +425,10 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
 
         case CHT_MODIFY:
         {
-            BOARD_ITEM* boardItemCopy = dynamic_cast<BOARD_ITEM*>( ent.m_copy );
+            BOARD_ITEM* boardItemCopy = nullptr;
+
+            if( ent.m_copy && ent.m_copy->IsBOARD_ITEM() )
+                boardItemCopy = static_cast<BOARD_ITEM*>( ent.m_copy );
 
             if( !( aCommitFlags & SKIP_UNDO ) )
             {
@@ -451,7 +446,7 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
                 connectivity->Update( boardItem );
             }
 
-            if( m_isBoardEditor && autofillZones )
+            if( m_isBoardEditor && autofillZones && boardItem->Type() != PCB_MARKER_T )
             {
                 dirtyIntersectingZones( boardItemCopy, changeType );   // before
                 dirtyIntersectingZones( boardItem, changeType );       // after
@@ -482,15 +477,6 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
                 } );
     }
 
-    if( bulkAddedItems.size() > 0 )
-        board->FinalizeBulkAdd( bulkAddedItems );
-
-    if( bulkRemovedItems.size() > 0 )
-        board->FinalizeBulkRemove( bulkRemovedItems );
-
-    if( itemsChanged.size() > 0 )
-        board->OnItemsChanged( itemsChanged );
-
     if( m_isBoardEditor )
     {
         size_t num_changes = m_changes.size();
@@ -519,14 +505,25 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
         }
 
         if( !staleTeardropPadsAndVias.empty() || !staleTeardropTracks.empty() )
+        {
             teardropMgr.UpdateTeardrops( *this, &staleTeardropPadsAndVias, &staleTeardropTracks );
 
+            // UpdateTeardrops() can modify the ratsnest data. So rebuild this ratsnest data
+            connectivity->RecalculateRatsnest( this );
+        }
+
         // Log undo items for any connectivity or teardrop changes
         for( size_t i = num_changes; i < m_changes.size(); ++i )
         {
             COMMIT_LINE& ent = m_changes[i];
-            BOARD_ITEM*  boardItem = dynamic_cast<BOARD_ITEM*>( ent.m_item );
-            BOARD_ITEM*  boardItemCopy = dynamic_cast<BOARD_ITEM*>( ent.m_copy );
+            BOARD_ITEM*  boardItem = nullptr;
+            BOARD_ITEM*  boardItemCopy = nullptr;
+
+            if( ent.m_item && ent.m_item->IsBOARD_ITEM() )
+                boardItem = static_cast<BOARD_ITEM*>( ent.m_item );
+
+            if( ent.m_copy && ent.m_copy->IsBOARD_ITEM() )
+                boardItemCopy = static_cast<BOARD_ITEM*>( ent.m_copy );
 
             wxCHECK2( boardItem, continue );
 
@@ -553,6 +550,9 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
         }
     }
 
+    if( bulkAddedItems.size() > 0 || bulkRemovedItems.size() > 0 || itemsChanged.size() > 0 )
+        board->OnItemsCompositeUpdate( bulkAddedItems, bulkRemovedItems, itemsChanged );
+
     if( frame )
     {
         if( !( aCommitFlags & SKIP_UNDO ) )
@@ -581,6 +581,15 @@ void BOARD_COMMIT::Push( const wxString& aMessage, int aCommitFlags )
             frame->OnModify();
         else
             frame->Update3DView( true, frame->GetPcbNewSettings()->m_Display.m_Live3DRefresh );
+
+        // Ensure the message panel is updated after committing changes.
+        // By default (i.e. if no event posted), display the updated board info
+        if( !itemsDeselected && !autofillZones && !selectedModified )
+        {
+            std::vector<MSG_PANEL_ITEM> msg_list;
+            board->GetMsgPanelInfo( frame, msg_list );
+            frame->SetMsgPanel( msg_list );
+        }
     }
 
     clear();
@@ -603,8 +612,8 @@ EDA_ITEM* BOARD_COMMIT::MakeImage( EDA_ITEM* aItem )
 {
     EDA_ITEM* clone = aItem->Clone();
 
-    if( BOARD_ITEM* board_item = dynamic_cast<BOARD_ITEM*>( clone ) )
-        board_item->SetParentGroup( nullptr );
+    if( clone->IsBOARD_ITEM() )
+        static_cast<BOARD_ITEM*>( clone )->SetParentGroup( nullptr );
 
     clone->SetFlags( UR_TRANSIENT );
 
@@ -628,12 +637,14 @@ void BOARD_COMMIT::Revert()
     for( auto it = m_changes.rbegin(); it != m_changes.rend(); ++it )
     {
         COMMIT_LINE& ent = *it;
-        BOARD_ITEM*  boardItem = dynamic_cast<BOARD_ITEM*>( ent.m_item );
+
+        if( !ent.m_item || !ent.m_item->IsBOARD_ITEM() )
+            continue;
+
+        BOARD_ITEM*  boardItem = static_cast<BOARD_ITEM*>( ent.m_item );
         int          changeType = ent.m_type & CHT_TYPE;
         int          changeFlags = ent.m_type & CHT_FLAGS;
 
-        wxCHECK2( boardItem, continue );
-
         switch( changeType )
         {
         case CHT_ADD:
@@ -664,15 +675,20 @@ void BOARD_COMMIT::Revert()
             break;
 
         case CHT_REMOVE:
+        {
             if( !( changeFlags & CHT_DONE ) )
                 break;
 
             view->Add( boardItem );
             connectivity->Add( boardItem );
 
-            if( FOOTPRINT* parentFP = dynamic_cast<FOOTPRINT*>( board->GetItem( ent.m_parent ) ) )
+            // Note: parent can be nullptr, because ent.m_parent is not
+            // initialized for every ent.m_item, only for some.
+            BOARD_ITEM* parent = board->GetItem( ent.m_parent );
+
+            if( parent && parent->Type() == PCB_FOOTPRINT_T )
             {
-                parentFP->Add( boardItem, ADD_MODE::INSERT );
+                static_cast<FOOTPRINT*>( parent )->Add( boardItem, ADD_MODE::INSERT );
             }
             else
             {
@@ -681,14 +697,15 @@ void BOARD_COMMIT::Revert()
             }
 
             break;
+        }
 
         case CHT_MODIFY:
         {
             view->Remove( boardItem );
             connectivity->Remove( boardItem );
 
-            BOARD_ITEM* boardItemCopy = dynamic_cast<BOARD_ITEM*>( ent.m_copy );
-            wxASSERT( boardItemCopy );
+            wxASSERT( ent.m_copy && ent.m_copy->IsBOARD_ITEM() );
+            BOARD_ITEM* boardItemCopy = static_cast<BOARD_ITEM*>( ent.m_copy );
             boardItem->SwapItemData( boardItemCopy );
 
             if( PCB_GROUP* group = dynamic_cast<PCB_GROUP*>( boardItem ) )
@@ -716,14 +733,8 @@ void BOARD_COMMIT::Revert()
         boardItem->ClearEditFlags();
     }
 
-    if( bulkAddedItems.size() > 0 )
-        board->FinalizeBulkAdd( bulkAddedItems );
-
-    if( bulkRemovedItems.size() > 0 )
-        board->FinalizeBulkRemove( bulkRemovedItems );
-
-    if( itemsChanged.size() > 0 )
-        board->OnItemsChanged( itemsChanged );
+    if( bulkAddedItems.size() > 0 || bulkRemovedItems.size() > 0 || itemsChanged.size() > 0 )
+        board->OnItemsCompositeUpdate( bulkAddedItems, bulkRemovedItems, itemsChanged );
 
     if( m_isBoardEditor )
     {
