@@ -1,7 +1,7 @@
 /*
  * This program source code file is part of KiCad, a free EDA CAD application.
  *
- * Copyright (C) 2023 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -61,19 +61,24 @@ protected:
 
     void build3DCanvas();
 
+    virtual void TearDownQuasiModal() override;
+
 protected:
     PANEL_FOOTPRINT_CHOOSER* m_chooserPanel;
-    bool    m_showFpMode;   // True to show the footprint, false for the 3D model
+    bool                     m_showFpMode;   // True to show the footprint, false for the 3D model
 
 private:
-    PCB_BASE_FRAME* m_parent;
-    BOARD_ADAPTER m_boardAdapter;
-    EDA_3D_CANVAS* m_preview3DCanvas;
-    CAMERA&     m_currentCamera;
-    TRACK_BALL  m_trackBallCamera;
-    BOARD*      m_dummyBoard;
-    BITMAP_BUTTON* m_grButtonFpView;
-    BITMAP_BUTTON* m_grButton3DView;
+    PCB_BASE_FRAME*          m_parent;
+    BOARD_ADAPTER            m_boardAdapter;
+    EDA_3D_CANVAS*           m_preview3DCanvas;
+    CAMERA&                  m_currentCamera;
+    TRACK_BALL               m_trackBallCamera;
+    BOARD*                   m_dummyBoard;
+    BITMAP_BUTTON*           m_grButtonFpView;
+    BITMAP_BUTTON*           m_grButton3DView;
+
+    /// The 3d viewer Render initial settings (must be saved and restored)
+    EDA_3D_VIEWER_SETTINGS::RENDER_SETTINGS m_initialRender;
 };
 
 #endif /* DIALOG_FOOTPRINT_CHOOSER_H */
