@@ -1,7 +1,7 @@
 /*
  * This program source code file is part of KiCad, a free EDA CAD application.
  *
- * Copyright (C) 2019-2023 KiCad Developers, see AUTHORS.txt for contributors.
+ * Copyright The KiCad Developers, see AUTHORS.txt for contributors.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -21,7 +21,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
-
+#include <algorithm>
 #include <cstdio>
 #include <memory>
 #include <mutex>
@@ -30,6 +30,7 @@
 #include <board_design_settings.h>
 #include <drc/drc_rtree.h>
 #include <drc/drc_engine.h>
+#include <lset.h>
 #include <pcb_track.h>
 #include <pcb_group.h>
 #include <geometry/shape_segment.h>
@@ -42,11 +43,10 @@
 bool fromToFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( aCtx ) : nullptr;
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
-
-    LIBEVAL::VALUE*   argTo = aCtx->Pop();
-    LIBEVAL::VALUE*   argFrom = aCtx->Pop();
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
+    LIBEVAL::VALUE*  argTo = aCtx->Pop();
+    LIBEVAL::VALUE*  argFrom = aCtx->Pop();
 
     result->Set(0.0);
     aCtx->Push( result );
@@ -77,10 +77,9 @@ bool fromToFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 static void existsOnLayerFunc( LIBEVAL::CONTEXT* aCtx, void *self )
 {
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( aCtx ) : nullptr;
-
-    LIBEVAL::VALUE*   arg = aCtx->Pop();
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
+    LIBEVAL::VALUE*  arg = aCtx->Pop();
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
 
     result->Set( 0.0 );
     aCtx->Push( result );
@@ -210,6 +209,36 @@ bool collidesWithCourtyard( BOARD_ITEM* aItem, std::shared_ptr<SHAPE>& aItemShap
 };
 
 
+static bool testFootprintSelector( FOOTPRINT* aFp, const wxString& aSelector )
+{
+    // NOTE: This code may want to be somewhat more generalized, but for now it's implemented
+    // here to support functions like insersectsCourtyard where we want multiple ways to search
+    // for the footprints in question.
+    // If support for text variable replacement is added, it should happen before any other
+    // logic here, so that people can use text variables to contain references or LIBIDs.
+    // (see: https://gitlab.com/kicad/code/kicad/-/issues/11231)
+
+    // First check if we have a known directive
+    if( aSelector.Upper().StartsWith( wxT( "${CLASS:" ) ) && aSelector.EndsWith( '}' ) )
+    {
+        wxString name = aSelector.Mid( 8, aSelector.Length() - 9 );
+
+        if( aFp->GetComponentClass()->ContainsClassName( name ) )
+            return true;
+    }
+    else if( aFp->GetReference().Matches( aSelector ) )
+    {
+        return true;
+    }
+    else if( aSelector.Contains( ':' ) && aFp->GetFPIDAsString().Matches( aSelector ) )
+    {
+        return true;
+    }
+
+    return false;
+}
+
+
 static bool searchFootprints( BOARD* aBoard, const wxString& aArg, PCBEXPR_CONTEXT* aCtx,
                               const std::function<bool( FOOTPRINT* )>& aFunc )
 {
@@ -229,11 +258,8 @@ static bool searchFootprints( BOARD* aBoard, const wxString& aArg, PCBEXPR_CONTE
     }
     else for( FOOTPRINT* fp : aBoard->Footprints() )
     {
-        if( fp->GetReference().Matches( aArg ) )
-        {
-            if( aFunc( fp ) )
-                return true;
-        }
+        if( testFootprintSelector( fp, aArg ) && aFunc( fp ) )
+            return true;
     }
 
     return false;
@@ -246,8 +272,8 @@ static bool searchFootprints( BOARD* aBoard, const wxString& aArg, PCBEXPR_CONTE
 static void intersectsCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_CONTEXT* context = static_cast<PCBEXPR_CONTEXT*>( aCtx );
-    LIBEVAL::VALUE*   arg = context->Pop();
-    LIBEVAL::VALUE*   result = context->AllocValue();
+    LIBEVAL::VALUE*  arg = context->Pop();
+    LIBEVAL::VALUE*  result = context->AllocValue();
 
     result->Set( 0.0 );
     context->Push( result );
@@ -261,7 +287,7 @@ static void intersectsCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( context ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( context ) : nullptr;
 
     if( !item )
         return;
@@ -310,8 +336,8 @@ static void intersectsCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 static void intersectsFrontCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_CONTEXT* context = static_cast<PCBEXPR_CONTEXT*>( aCtx );
-    LIBEVAL::VALUE*   arg = context->Pop();
-    LIBEVAL::VALUE*   result = context->AllocValue();
+    LIBEVAL::VALUE*  arg = context->Pop();
+    LIBEVAL::VALUE*  result = context->AllocValue();
 
     result->Set( 0.0 );
     context->Push( result );
@@ -325,7 +351,7 @@ static void intersectsFrontCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( context ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( context ) : nullptr;
 
     if( !item )
         return;
@@ -373,8 +399,8 @@ static void intersectsFrontCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 static void intersectsBackCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_CONTEXT* context = static_cast<PCBEXPR_CONTEXT*>( aCtx );
-    LIBEVAL::VALUE*   arg = context->Pop();
-    LIBEVAL::VALUE*   result = context->AllocValue();
+    LIBEVAL::VALUE*  arg = context->Pop();
+    LIBEVAL::VALUE*  result = context->AllocValue();
 
     result->Set( 0.0 );
     context->Push( result );
@@ -388,7 +414,7 @@ static void intersectsBackCourtyardFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( context ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( context ) : nullptr;
 
     if( !item )
         return;
@@ -524,12 +550,17 @@ bool collidesWithArea( BOARD_ITEM* aItem, PCBEXPR_CONTEXT* aCtx, ZONE* aArea )
 
         if( zoneRTree )
         {
-            for( PCB_LAYER_ID layer : aArea->GetLayerSet().Seq() )
+            for( size_t ii = 0; ii < aArea->GetLayerSet().size(); ++ii )
             {
-                if( aCtx->GetLayer() == layer || aCtx->GetLayer() == UNDEFINED_LAYER )
+                if( aArea->GetLayerSet().test( ii ) )
                 {
-                    if( zoneRTree->QueryColliding( areaBBox, &areaOutline, layer ) )
-                        return true;
+                    PCB_LAYER_ID layer = PCB_LAYER_ID( ii );
+
+                    if( aCtx->GetLayer() == layer || aCtx->GetLayer() == UNDEFINED_LAYER )
+                    {
+                        if( zoneRTree->QueryColliding( areaBBox, &areaOutline, layer ) )
+                            return true;
+                    }
                 }
             }
         }
@@ -617,14 +648,39 @@ bool searchAreas( BOARD* aBoard, const wxString& aArg, PCBEXPR_CONTEXT* aCtx,
 }
 
 
+class SCOPED_LAYERSET
+{
+public:
+    SCOPED_LAYERSET( BOARD_ITEM* aItem )
+    {
+        m_item = aItem;
+        m_layers = aItem->GetLayerSet();
+    }
+
+    ~SCOPED_LAYERSET()
+    {
+        m_item->SetLayerSet( m_layers );
+    }
+
+    void Add( PCB_LAYER_ID aLayer )
+    {
+        m_item->SetLayerSet( m_item->GetLayerSet().set( aLayer ) );
+    }
+
+private:
+    BOARD_ITEM* m_item;
+    LSET        m_layers;
+};
+
+
 #define MISSING_AREA_ARG( f ) \
     wxString::Format( _( "Missing rule-area argument (A, B, or rule-area name) to %s." ), f )
 
 static void intersectsAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_CONTEXT* context = static_cast<PCBEXPR_CONTEXT*>( aCtx );
-    LIBEVAL::VALUE*   arg = aCtx->Pop();
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
+    LIBEVAL::VALUE*  arg = aCtx->Pop();
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
 
     result->Set( 0.0 );
     aCtx->Push( result );
@@ -638,7 +694,7 @@ static void intersectsAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( context ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( context ) : nullptr;
 
     if( !item )
         return;
@@ -656,6 +712,18 @@ static void intersectsAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
                             if( !aArea || aArea == item || aArea->GetParent() == item )
                                 return false;
 
+                            SCOPED_LAYERSET scopedLayerSet( aArea );
+
+                            if( context->GetConstraint() == SILK_CLEARANCE_CONSTRAINT )
+                            {
+                                // Silk clearance tests are run across layer pairs
+                                if(    ( aArea->IsOnLayer( F_SilkS ) && IsFrontLayer( aLayer ) )
+                                    || ( aArea->IsOnLayer( B_SilkS ) && IsBackLayer( aLayer ) ) )
+                                {
+                                    scopedLayerSet.Add( aLayer );
+                                }
+                            }
+
                             LSET commonLayers = aArea->GetLayerSet() & item->GetLayerSet();
 
                             if( !commonLayers.any() )
@@ -711,8 +779,8 @@ static void intersectsAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 static void enclosedByAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_CONTEXT* context = static_cast<PCBEXPR_CONTEXT*>( aCtx );
-    LIBEVAL::VALUE*   arg = aCtx->Pop();
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
+    LIBEVAL::VALUE*  arg = aCtx->Pop();
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
 
     result->Set( 0.0 );
     aCtx->Push( result );
@@ -726,7 +794,7 @@ static void enclosedByAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( context ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( context ) : nullptr;
 
     if( !item )
         return;
@@ -745,8 +813,11 @@ static void enclosedByAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
                             if( !aArea || aArea == item || aArea->GetParent() == item )
                                 return false;
 
-                            if( !( aArea->GetLayerSet() & item->GetLayerSet() ).any() )
-                                return false;
+                            if( item->Type() != PCB_FOOTPRINT_T )
+                            {
+                                if( !( aArea->GetLayerSet() & item->GetLayerSet() ).any() )
+                                    return false;
+                            }
 
                             if( !aArea->GetBoundingBox().Intersects( itemBBox ) )
                                 return false;
@@ -766,8 +837,27 @@ static void enclosedByAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
                             SHAPE_POLY_SET itemShape;
                             bool           enclosedByArea;
 
-                            item->TransformShapeToPolygon( itemShape, layer, 0, maxError,
-                                                           ERROR_OUTSIDE );
+                            if( item->Type() == PCB_ZONE_T )
+                            {
+                                itemShape = *static_cast<ZONE*>( item )->Outline();
+                            }
+                            else if( item->Type() == PCB_FOOTPRINT_T )
+                            {
+                                FOOTPRINT* fp = static_cast<FOOTPRINT*>( item );
+
+                                for( PCB_LAYER_ID testLayer : aArea->GetLayerSet() )
+                                {
+                                    fp->TransformPadsToPolySet( itemShape, testLayer, 0,
+                                                                maxError, ERROR_OUTSIDE );
+                                    fp->TransformFPShapesToPolySet( itemShape, testLayer, 0,
+                                                                    maxError, ERROR_OUTSIDE );
+                                }
+                            }
+                            else
+                            {
+                                item->TransformShapeToPolygon( itemShape, layer, 0, maxError,
+                                                               ERROR_OUTSIDE );
+                            }
 
                             if( itemShape.IsEmpty() )
                             {
@@ -776,8 +866,7 @@ static void enclosedByAreaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
                             }
                             else
                             {
-                                itemShape.BooleanSubtract( *aArea->Outline(),
-                                                           SHAPE_POLY_SET::PM_FAST );
+                                itemShape.BooleanSubtract( *aArea->Outline() );
 
                                 enclosedByArea = itemShape.IsEmpty();
                             }
@@ -819,7 +908,7 @@ static void memberOfGroupFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( aCtx ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
 
     if( !item )
         return;
@@ -865,7 +954,7 @@ static void memberOfSheetFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( aCtx ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
 
     if( !item )
         return;
@@ -875,14 +964,25 @@ static void memberOfSheetFunc( LIBEVAL::CONTEXT* aCtx, void* self )
             {
                 FOOTPRINT* fp = item->GetParentFootprint();
 
+                if( !fp && item->Type() == PCB_FOOTPRINT_T )
+                    fp = static_cast<FOOTPRINT*>( item );
+
                 if( !fp )
                     return 0.0;
 
-                if( fp->GetSheetname().Matches( arg->AsString() ) )
+                wxString sheetName = fp->GetSheetname();
+                wxString refName = arg->AsString();
+
+                if( sheetName.EndsWith( wxT("/") ) )
+                    sheetName.RemoveLast();
+                if( refName.EndsWith( wxT("/") ) )
+                    refName.RemoveLast();
+
+                if( sheetName.Matches( refName ) )
                     return 1.0;
 
-                if( ( arg->AsString().Matches( wxT( "/" ) ) || arg->AsString().IsEmpty() )
-                    && fp->GetSheetname().IsEmpty() )
+                if( ( refName.Matches( wxT( "/" ) ) || refName.IsEmpty() )
+                    && sheetName.IsEmpty() )
                 {
                     return 1.0;
                 }
@@ -912,7 +1012,7 @@ static void memberOfFootprintFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     }
 
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( aCtx ) : nullptr;
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
 
     if( !item )
         return;
@@ -920,18 +1020,10 @@ static void memberOfFootprintFunc( LIBEVAL::CONTEXT* aCtx, void* self )
     result->SetDeferredEval(
             [item, arg]() -> double
             {
-                ;
-
                 if( FOOTPRINT* parentFP = item->GetParentFootprint() )
                 {
-                    if( parentFP->GetReference().Matches( arg->AsString() ) )
+                    if( testFootprintSelector( parentFP, arg->AsString() ) )
                         return 1.0;
-
-                    if( arg->AsString().Contains( ':' )
-                             && parentFP->GetFPIDAsString().Matches( arg->AsString() ) )
-                    {
-                        return 1.0;
-                    }
                 }
 
                 return 0.0;
@@ -942,8 +1034,8 @@ static void memberOfFootprintFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 static void isMicroVia( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( aCtx ) : nullptr;
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
 
     result->Set( 0.0 );
     aCtx->Push( result );
@@ -959,8 +1051,8 @@ static void isMicroVia( LIBEVAL::CONTEXT* aCtx, void* self )
 static void isBlindBuriedViaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
     PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item = vref ? vref->GetObject( aCtx ) : nullptr;
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
 
     result->Set( 0.0 );
     aCtx->Push( result );
@@ -975,7 +1067,7 @@ static void isBlindBuriedViaFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 
 static void isCoupledDiffPairFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
-    PCBEXPR_CONTEXT*     context = static_cast<PCBEXPR_CONTEXT*>( aCtx );
+    PCBEXPR_CONTEXT*      context = static_cast<PCBEXPR_CONTEXT*>( aCtx );
     BOARD_CONNECTED_ITEM* a = dynamic_cast<BOARD_CONNECTED_ITEM*>( context->GetItem( 0 ) );
     BOARD_CONNECTED_ITEM* b = dynamic_cast<BOARD_CONNECTED_ITEM*>( context->GetItem( 1 ) );
     LIBEVAL::VALUE*       result = aCtx->AllocValue();
@@ -1014,10 +1106,10 @@ static void isCoupledDiffPairFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 
 static void inDiffPairFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
-    LIBEVAL::VALUE*   argv   = aCtx->Pop();
+    LIBEVAL::VALUE*  argv   = aCtx->Pop();
     PCBEXPR_VAR_REF* vref   = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item   = vref ? vref->GetObject( aCtx ) : nullptr;
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
+    BOARD_ITEM*      item   = vref ? vref->GetObject( aCtx ) : nullptr;
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
 
     result->Set( 0.0 );
     aCtx->Push( result );
@@ -1065,10 +1157,10 @@ static void inDiffPairFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 
 static void getFieldFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 {
-    LIBEVAL::VALUE*   arg    = aCtx->Pop();
+    LIBEVAL::VALUE*  arg    = aCtx->Pop();
     PCBEXPR_VAR_REF* vref   = static_cast<PCBEXPR_VAR_REF*>( self );
-    BOARD_ITEM*       item   = vref ? vref->GetObject( aCtx ) : nullptr;
-    LIBEVAL::VALUE*   result = aCtx->AllocValue();
+    BOARD_ITEM*      item   = vref ? vref->GetObject( aCtx ) : nullptr;
+    LIBEVAL::VALUE*  result = aCtx->AllocValue();
 
     result->Set( "" );
     aCtx->Push( result );
@@ -1105,6 +1197,85 @@ static void getFieldFunc( LIBEVAL::CONTEXT* aCtx, void* self )
 }
 
 
+static void hasNetclassFunc( LIBEVAL::CONTEXT* aCtx, void* self )
+{
+    LIBEVAL::VALUE* arg = aCtx->Pop();
+    LIBEVAL::VALUE* result = aCtx->AllocValue();
+
+    result->Set( 0.0 );
+    aCtx->Push( result );
+
+    if( !arg || arg->AsString().IsEmpty() )
+    {
+        if( aCtx->HasErrorCallback() )
+            aCtx->ReportError( _( "Missing netclass name argument to hasNetclass()" ) );
+
+        return;
+    }
+
+    PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
+
+    if( !item )
+        return;
+
+    result->SetDeferredEval(
+            [item, arg]() -> double
+            {
+                if( !item->IsConnected() )
+                    return 0.0;
+
+                BOARD_CONNECTED_ITEM* bcItem = static_cast<BOARD_CONNECTED_ITEM*>( item );
+                NETCLASS*             netclass = bcItem->GetEffectiveNetClass();
+
+                if( netclass->ContainsNetclassWithName( arg->AsString() ) )
+                    return 1.0;
+
+                return 0.0;
+            } );
+}
+
+static void hasComponentClassFunc( LIBEVAL::CONTEXT* aCtx, void* self )
+{
+    LIBEVAL::VALUE* arg = aCtx->Pop();
+    LIBEVAL::VALUE* result = aCtx->AllocValue();
+
+    result->Set( 0.0 );
+    aCtx->Push( result );
+
+    if( !arg || arg->AsString().IsEmpty() )
+    {
+        if( aCtx->HasErrorCallback() )
+            aCtx->ReportError(
+                    _( "Missing component class name argument to hasComponentClass()" ) );
+
+        return;
+    }
+
+    PCBEXPR_VAR_REF* vref = static_cast<PCBEXPR_VAR_REF*>( self );
+    BOARD_ITEM*      item = vref ? vref->GetObject( aCtx ) : nullptr;
+
+    if( !item )
+        return;
+
+    result->SetDeferredEval(
+            [item, arg]() -> double
+            {
+                FOOTPRINT* footprint = dynamic_cast<FOOTPRINT*>( item );
+
+                if( !footprint )
+                    return 0.0;
+
+                const COMPONENT_CLASS* compClass = footprint->GetComponentClass();
+
+                if( compClass && compClass->ContainsClassName( arg->AsString() ) )
+                    return 1.0;
+
+                return 0.0;
+            } );
+}
+
+
 PCBEXPR_BUILTIN_FUNCTIONS::PCBEXPR_BUILTIN_FUNCTIONS()
 {
     RegisterAllFunctions();
@@ -1143,6 +1314,7 @@ void PCBEXPR_BUILTIN_FUNCTIONS::RegisterAllFunctions()
     RegisterFunc( wxT( "inDiffPair('x')" ), inDiffPairFunc );
 
     RegisterFunc( wxT( "getField('x')" ), getFieldFunc );
+
+    RegisterFunc( wxT( "hasNetclass('x')" ), hasNetclassFunc );
+    RegisterFunc( wxT( "hasComponentClass('x')" ), hasComponentClassFunc );
 }
-
-
